@@ -1,9 +1,9 @@
 #!/usr/bin/python2
 # -*- coding: utf-8 -*-
 #
-# Automatic patch dependency checker and apply script/README.md generator.
+# Automatic patch dependency checker and apply script generator.
 #
-# Copyright (C) 2014-2015 Sebastian Lackner
+# Copyright (C) 2014-2016 Sebastian Lackner
 # Copyright (C) 2015 Michael Müller
 #
 # This library is free software; you can redistribute it and/or
@@ -46,29 +46,24 @@ import ConfigParser
 _devnull = open(os.devnull, 'wb')
 
 # Cached information to speed up patch dependency checks
-latest_wine_commit     = None
-latest_staging_version = None
-cached_patch_result    = {}
+upstream_commit = None
 
 class config(object):
     path_cache              = ".patchupdate.cache"
     path_config             = os.path.expanduser("~/.config/patchupdate.conf")
 
     path_patches            = "patches"
-    path_changelog          = "staging/changelog"
+    path_version            = "staging/VERSION"
     path_wine               = "staging/wine"
 
     path_template_script    = "staging/patchinstall.sh.in"
     path_script             = "patches/patchinstall.sh"
 
-    path_template_README_md = "staging/README.md.in"
-    path_README_md          = "README.md"
-
     path_IfDefined          = "9999-IfDefined.patch"
 
     bugtracker_url          = "https://bugs.winehq.org/xmlrpc.cgi"
     bugtracker_defaultcc    = ["michael@fds-team.de", "sebastian@fds-team.de",
-                               "erich.e.hoover@wine-staging.com"]
+                               "erich.e.hoover@wine-staging.com", "dmitry@baikal.ru"]
     bugtracker_user         = None
     bugtracker_pass         = None
 
@@ -148,31 +143,19 @@ def _sha256(fp):
 
 def _parse_int(val, default=0):
     """Parse an integer or boolean value."""
-    r = re.match("^[0-9]+$", val)
-    if r:
+    if re.match("^[0-9]+$", val):
         return int(val)
     try:
         return {'true': 1, 'yes': 1, 'false': 0, 'no': 0}[val.lower()]
-    except AttributeError:
+    except KeyError:
         return default
 
-def _read_changelog():
-    """Read information from changelog."""
-    with open(config.path_changelog) as fp:
-        for line in fp:
-            r = re.match("^([a-zA-Z0-9][^(]*)\((.*)\) ([^;]*)", line)
-            if r: yield (r.group(1).strip(), r.group(2).strip(), r.group(3).strip())
+def _staging_version():
+    """Get the current version number of Wine Staging."""
+    with open(config.path_version) as fp:
+        return fp.read().strip()
 
-def _latest_staging_version(only_released=False):
-    """Get version number of the latest release / unreleased version."""
-    for package, version, distro in _read_changelog():
-        version = version.replace("~rc", "-rc")
-        if distro.lower() != "unreleased":
-            return version
-        elif not only_released:
-            return "%s (unreleased)" % version
-
-def _latest_wine_commit(commit=None):
+def _upstream_commit(commit=None):
     """Get latest wine commit."""
     if not os.path.isdir(config.path_wine):
         raise PatchUpdaterError("Please create a symlink to the wine repository in %s" % config.path_wine)
@@ -181,73 +164,43 @@ def _latest_wine_commit(commit=None):
     assert len(commit) == 40 and commit == commit.lower()
     return commit
 
-def enum_directories(revision, path):
-    """Enumerate all subdirectories of 'path' at a specific revision."""
+def enum_patchsets(path):
+    """Return a sorted list of all subdirectories of path."""
     dirs = []
+    for name in os.listdir(path):
+        directory = os.path.join(path, name)
+        if not os.path.isdir(directory):
+            continue
+        dirs.append((name, directory))
+    return sorted(dirs)
 
-    if path[0:2] == "./":
-        path = path[2:]
-    elif path[0] == "/":
-        raise RuntimeError("Expected relative path, not an absolute path")
-
-    if revision is None:
-        for name in os.listdir(path):
-            if name in [".", ".."]: continue
-            directory = os.path.join(path, name)
-            if not os.path.isdir(directory):
-                continue
-            dirs.append((name, directory))
-    else:
-        filename = "%s:%s" % (revision, path)
-        try:
-            content = subprocess.check_output(["git", "show", filename], stderr=_devnull)
-        except subprocess.CalledProcessError as e:
-            if e.returncode != 128: raise
-            return [] # ignore error
-        lines = content.split("\n")
-        if not lines[0].startswith("tree ") or lines[1] != "":
-            raise RuntimeError("Unexpected output from 'git show %s'" % filename)
-        for name in lines[2:]:
-            if name == "" or name[-1] != "/": continue
-            name = name[:-1]
-            dirs.append((name, os.path.join(path, name)))
-
-    return dirs
-
-def read_patchset(revision = None):
-    """Read information about all patchsets for a specific revision."""
+def load_patchsets():
+    """Read information about all patchsets."""
     unique_id   = itertools.count()
     all_patches = {}
     name_to_id  = {}
     categories  = {}
 
-    # Read in sorted order (to ensure created Makefile doesn't change too much)
-    for name, directory in sorted(enum_directories(revision, config.path_patches)):
+    for name, directory in enum_patchsets(config.path_patches):
         patch = PatchSet(name, directory)
 
-        if revision is None:
-
-            # If its the latest revision, then request additional information
-            if not os.path.isdir(directory):
-                raise RuntimeError("Unable to open directory %s" % directory)
-
-            # Enumerate .patch files in the given directory, enumerate individual patches and affected files
-            for f in sorted(os.listdir(directory)):
-                if not re.match("^[0-9]{4}-.*\\.patch$", f):
-                    continue
-                if f.startswith(config.path_IfDefined):
-                    continue
-                if not os.path.isfile(os.path.join(directory, f)):
-                    continue
-                patch.files.append(f)
-                for p in patchutils.read_patch(os.path.join(directory, f)):
-                    patch.modified_files.add(p.modified_file)
-                    patch.patches.append(p)
-
-            # No single patch within this directory, ignore it
-            if len(patch.patches) == 0:
-                del patch
+        # Enumerate .patch files in the given directory, enumerate individual patches and affected files
+        for f in sorted(os.listdir(directory)):
+            if not re.match("^[0-9]{4}-.*\\.patch$", f):
                 continue
+            if f.startswith(config.path_IfDefined):
+                continue
+            if not os.path.isfile(os.path.join(directory, f)):
+                continue
+            patch.files.append(f)
+            for p in patchutils.read_patch(os.path.join(directory, f)):
+                patch.modified_files.add(p.modified_file)
+                patch.patches.append(p)
+
+        # No single patch within this directory, ignore it
+        if len(patch.patches) == 0:
+            del patch
+            continue
 
         i = next(unique_id)
         all_patches[i]   = patch
@@ -255,15 +208,11 @@ def read_patchset(revision = None):
 
     # Now read the definition files in a second step
     for i, patch in all_patches.iteritems():
+        filename = os.path.join(os.path.join(config.path_patches, patch.name), "definition")
         try:
-            filename = os.path.join(os.path.join(config.path_patches, patch.name), "definition")
-            if revision is None:
-                with open(filename) as fp:
-                    content = fp.read()
-            else:
-                filename = "%s:%s" % (revision, filename)
-                content = subprocess.check_output(["git", "show", filename], stderr=_devnull)
-        except (IOError, subprocess.CalledProcessError):
+            with open(filename) as fp:
+                content = fp.read()
+        except IOError:
             continue # Skip this definition file
 
         for line in content.split("\n"):
@@ -310,16 +259,21 @@ def read_patchset(revision = None):
             elif key == "ifdefined":
                 patch.ifdefined = val
 
-            elif revision is None:
+            else:
                 print "WARNING: Ignoring unknown command in definition file %s: %s" % (filename, line)
 
         # If patch is not disabled then finally add it to the category
-        if not patch.disabled:
-            for category in patch.categories:
-                if not categories.has_key(category):
-                    categories[category] = set()
-                categories[category].add(i)
+        if patch.disabled:
+            continue
 
+        for category in patch.categories:
+            if not categories.has_key(category):
+                categories[category] = set()
+            categories[category].add(i)
+
+    # Filter autodepends on disabled patchsets
+    for i, patch in all_patches.iteritems():
+        patch.auto_depends = set([j for j in patch.auto_depends if not all_patches[j].disabled])
 
     # Add virtual targets for all the categories
     for category, indices in categories.iteritems():
@@ -360,13 +314,13 @@ def contains_binary_patch(all_patches, indices, filename):
     """Checks if any patch with given indices affecting filename is a binary patch."""
     for i in indices:
         for patch in all_patches[i].patches:
-            if patch.modified_file == filename and patch.is_binary():
+            if patch.modified_file == filename and patch.is_binary:
                 return True
     return False
 
 def get_wine_file(filename):
     """Return the content of a file."""
-    entry  = "%s:%s" % (latest_wine_commit, filename)
+    entry  = "%s:%s" % (upstream_commit, filename)
     result = tempfile.NamedTemporaryFile()
     try:
         content = subprocess.check_call(["git", "show", entry], cwd=config.path_wine, \
@@ -405,14 +359,11 @@ def resolve_dependencies(all_patches, index = None, depends = None, auto_deps =
 
     def _resolve(depends):
         for i in sorted(depends):
-            # Check for disabled patch
-            if all_patches[i].disabled:
+            if all_patches[i].disabled: # Check for disabled patch
                 raise PatchUpdaterError("Encountered dependency on disabled patchset %s" % all_patches[i].name)
-            # Dependencies already resolved
-            if all_patches[i].verify_resolved > 0:
+            if all_patches[i].verify_resolved > 0: # Dependencies already resolved
                 continue
-            # Detect circular dependency
-            if all_patches[i].verify_resolved < 0:
+            if all_patches[i].verify_resolved < 0: # Detect circular dependency
                 raise PatchUpdaterError("Circular dependency while trying to resolve %s" % all_patches[i].name)
 
             # Recusively resolve dependencies
@@ -532,7 +483,7 @@ def generate_ifdefined(all_patches, skip_checks=False):
             fp.write("\n")
 
             depends = resolve_dependencies(enabled_patches, i)
-            for f in patch.modified_files:
+            for f in sorted(patch.modified_files):
 
                 # Reconstruct the state after applying the dependencies
                 original = get_wine_file(f)
@@ -581,115 +532,127 @@ def generate_ifdefined(all_patches, skip_checks=False):
             assert p.modified_file in patch.modified_files
             patch.patches.append(p)
 
-def generate_script(all_patches, skip_checks=False):
+def generate_apply_order(all_patches, skip_checks=False):
     """Resolve dependencies, and afterwards check if everything applies properly."""
     depends     = sorted([i for i, patch in all_patches.iteritems() if not patch.disabled])
     resolved    = resolve_dependencies(all_patches, depends=depends)
     max_patches = max(resolved) + 1
 
-    if not skip_checks:
+    if skip_checks:
+        return resolved
 
-        # Generate timestamps based on dependencies, still required for binary patches
-        # Find out which files are modified by multiple patches
-        modified_files = {}
-        for i, patch in [(i, all_patches[i]) for i in resolved]:
-            patch.verify_time = [0]*max_patches
-            patch.verify_time[i] += 1
-            for j in patch.depends:
-                patch.verify_time = causal_time_combine(patch.verify_time, all_patches[j].verify_time)
+    # Generate timestamps based on dependencies, still required for binary patches
+    # Find out which files are modified by multiple patches
+    modified_files = {}
+    for i, patch in [(i, all_patches[i]) for i in resolved]:
+        patch.verify_time = [0]*max_patches
+        patch.verify_time[i] += 1
+        for j in patch.depends:
+            patch.verify_time = causal_time_combine(patch.verify_time, all_patches[j].verify_time)
 
-            for f in patch.modified_files:
-                if f not in modified_files:
-                    modified_files[f] = []
-                modified_files[f].append(i)
+        for f in patch.modified_files:
+            if f not in modified_files:
+                modified_files[f] = []
+            modified_files[f].append(i)
 
-        # Check dependencies
-        dependency_cache = _load_dict(config.path_cache)
-        pool = multiprocessing.pool.ThreadPool(processes=4)
-        try:
-            for filename, indices in modified_files.iteritems():
+    # Check dependencies
+    dependency_cache = _load_dict(config.path_cache)
+    pool = multiprocessing.pool.ThreadPool(processes=4)
+    try:
+        for filename, indices in modified_files.iteritems():
 
-                # If one of patches is a binary patch, then we cannot / won't verify it - require dependencies in this case
-                if contains_binary_patch(all_patches, indices, filename):
-                    if not causal_time_relation_any(all_patches, indices):
-                        raise PatchUpdaterError("Because of binary patch modifying file %s the following patches need explicit dependencies: %s" %
-                                                (filename, ", ".join([all_patches[i].name for i in indices])))
+            # If one of patches is a binary patch, then we cannot / won't verify it - require dependencies in this case
+            if contains_binary_patch(all_patches, indices, filename):
+                if not causal_time_relation_any(all_patches, indices):
+                    raise PatchUpdaterError("Because of binary patch modifying file %s the following patches need explicit dependencies: %s" %
+                                            (filename, ", ".join([all_patches[i].name for i in indices])))
+                continue
+
+            original_content = get_wine_file(filename)
+            original_hash    = _sha256(original_content)
+            selected_patches = select_patches(all_patches, indices, filename)
+
+            # Generate a unique id based on the original content, the selected patches
+            # and the dependency information. Since this information only has to be compared
+            # we can throw it into a single hash.
+            m = hashlib.sha256()
+            m.update(original_hash)
+            for i in indices:
+                m.update("P%s" % selected_patches[i][0])
+                for j in indices:
+                    if causal_time_smaller(all_patches[j].verify_time, all_patches[i].verify_time):
+                        m.update("D%s" % selected_patches[j][0])
+            unique_hash = m.digest()
+
+            # Skip checks if it matches the information from the cache
+            # For backwards compatibility, convert string entries to list
+            if dependency_cache.has_key(filename):
+                if not isinstance(dependency_cache[filename], list):
+                    dependency_cache[filename] = [dependency_cache[filename]]
+                if unique_hash in dependency_cache[filename]:
+                    dependency_cache[filename].append(unique_hash)
+                    dependency_cache[filename].remove(unique_hash)
                     continue
 
-                original_content = get_wine_file(filename)
-                original_hash    = _sha256(original_content)
-                selected_patches = select_patches(all_patches, indices, filename)
+            # Show a progress bar while applying the patches - this task might take some time
+            chunk_size = 20
+            with progressbar.ProgressBar(desc=filename, total=2 ** len(indices) / chunk_size) as progress:
 
-                # Generate a unique id based on the original content, the selected patches
-                # and the dependency information. Since this information only has to be compared
-                # we can throw it into a single hash.
-                m = hashlib.sha256()
-                m.update(original_hash)
-                for i in indices:
-                    m.update("P%s" % selected_patches[i][0])
-                    for j in indices:
-                        if causal_time_smaller(all_patches[j].verify_time, all_patches[i].verify_time):
-                            m.update("D%s" % selected_patches[j][0])
-                unique_hash = m.digest()
+                def test_apply(current):
+                    set_apply = [(i, all_patches[i]) for i in current]
+                    set_skip  = [(i, all_patches[i]) for i in indices if i not in current]
 
-                # Skip checks if it matches the information from the cache
-                try:
-                    if dependency_cache[filename] == unique_hash:
-                        continue
-                except KeyError:
-                    pass
+                    # Check if there is any patch2 which depends directly or indirectly on patch1.
+                    # If this is the case we found an impossible situation, we can be skipped in this test.
+                    for i, patch1 in set_apply:
+                        for j, patch2 in set_skip:
+                            if causal_time_smaller(patch2.verify_time, patch1.verify_time):
+                                return True # we can skip this test
 
-                # Show a progress bar while applying the patches - this task might take some time
-                chunk_size = 20
-                with progressbar.ProgressBar(desc=filename, total=2 ** len(indices) / chunk_size) as progress:
+                    try:
+                        original = original_content
+                        for i, _ in set_apply:
+                            original = patchutils.apply_patch(original, selected_patches[i][1], fuzz=0)
+                    except patchutils.PatchApplyError:
+                        return False
 
-                    def test_apply(current):
-                        set_apply = [(i, all_patches[i]) for i in current]
-                        set_skip  = [(i, all_patches[i]) for i in indices if i not in current]
+                    return True # everything is fine
 
-                        # Check if there is any patch2 which depends directly or indirectly on patch1.
-                        # If this is the case we found an impossible situation, we can be skipped in this test.
-                        for i, patch1 in set_apply:
-                            for j, patch2 in set_skip:
-                                if causal_time_smaller(patch2.verify_time, patch1.verify_time):
-                                    return True # we can skip this test
+                def test_apply_seq(current_list):
+                    for current in current_list:
+                        if not test_apply(current):
+                            return current
+                    return None
 
-                        try:
-                            original = original_content
-                            for i, _ in set_apply:
-                                original = patchutils.apply_patch(original, selected_patches[i][1], fuzz=0)
-                        except patchutils.PatchApplyError:
-                            return False
+                iterables = []
+                for i in xrange(0, len(indices) + 1):
+                    iterables.append(itertools.combinations(indices, i))
+                it = _split_seq(itertools.chain(*iterables), chunk_size)
+                for k, failed in enumerate(pool.imap_unordered(test_apply_seq, it)):
+                    if failed is not None:
+                        progress.finish("<failed to apply>")
+                        raise PatchUpdaterError("Changes to file %s don't apply: %s" %
+                                                (filename, ", ".join([all_patches[i].name for i in failed])))
+                    progress.update(k)
 
-                        return True # everything is fine
+            # Update the dependency cache, store max 10 entries per file
+            if not dependency_cache.has_key(filename):
+                dependency_cache[filename] = []
+            dependency_cache[filename].append(unique_hash)
+            dependency_cache[filename] = dependency_cache[filename][-10:]
 
-                    def test_apply_seq(current_list):
-                        for current in current_list:
-                            if not test_apply(current):
-                                return current
-                        return None
+        # Delete outdated cache information
+        for filename in dependency_cache.keys():
+            if not modified_files.has_key(filename):
+                del dependency_cache[filename]
+    finally:
+        pool.close()
+        _save_dict(config.path_cache, dependency_cache)
 
-                    iterables = []
-                    for i in xrange(0, len(indices) + 1):
-                        iterables.append(itertools.combinations(indices, i))
-                    it = _split_seq(itertools.chain(*iterables), chunk_size)
-                    for k, failed in enumerate(pool.imap_unordered(test_apply_seq, it)):
-                        if failed is not None:
-                            progress.finish("<failed to apply>")
-                            raise PatchUpdaterError("Changes to file %s don't apply: %s" %
-                                                    (filename, ", ".join([all_patches[i].name for i in failed])))
-                        progress.update(k)
+    return resolved
 
-                # Update the dependency cache
-                dependency_cache[filename] = unique_hash
-
-            # Delete outdated cache information
-            for filename in dependency_cache.keys():
-                if not modified_files.has_key(filename):
-                    del dependency_cache[filename]
-        finally:
-            pool.close()
-            _save_dict(config.path_cache, dependency_cache)
+def generate_script(all_patches, resolved):
+    """Generate script to apply patches."""
 
     # Generate code for helper functions
     lines = []
@@ -790,8 +753,8 @@ def generate_script(all_patches, skip_checks=False):
     with open(config.path_template_script) as template_fp:
         template = template_fp.read()
     with open(config.path_script, "w") as fp:
-        fp.write(template.format(latest_staging_version=_latest_staging_version(),
-                                 latest_wine_commit=latest_wine_commit,
+        fp.write(template.format(staging_version=_staging_version(),
+                                 upstream_commit=upstream_commit,
                                  patch_helpers="".join(lines_helpers).rstrip("\n"),
                                  patch_resolver="".join(lines_resolver).rstrip("\n"),
                                  patch_apply="".join(lines_apply).rstrip("\n")))
@@ -799,100 +762,6 @@ def generate_script(all_patches, skip_checks=False):
     # Add changes to git
     subprocess.call(["git", "add", config.path_script])
 
-def generate_markdown(all_patches, release_patches):
-    """Generate README.md including information about specific patches and bugfixes."""
-
-    def _format_bug(mode, bugid, bugname):
-        if mode < 0: bugname = "~~%s~~" % bugname
-        if bugid is None: return "* %s" % bugname
-        return "* %s ([Wine Bug #%d](https://bugs.winehq.org/show_bug.cgi?id=%d))" % \
-               (bugname, bugid, bugid) #, short_desc.replace("\\", "\\\\").replace("\"", "\\\""))
-
-    all_fixes = {}
-
-    # Get fixes for current version
-    for _, patch in all_patches.iteritems():
-        for sync, bugid, bugname in patch.fixes:
-            key = bugid if bugid is not None else bugname
-            all_fixes[key] = [1, bugid, bugname]
-
-    # Compare with fixes for last release
-    for _, patch in release_patches.iteritems():
-        for sync, bugid, bugname in patch.fixes:
-            if bugid is not None and all_fixes.has_key(bugid):
-                all_fixes[bugid][0] = 0
-            elif all_fixes.has_key(bugname):
-                all_fixes[bugname][0] = 0
-            elif bugid is None:
-                for k, v in all_fixes.iteritems():
-                    if v[2] != bugname: continue
-                    if v[1] is None: continue
-                    all_fixes[v[1]][0] = 0
-                    break
-                else:
-                    all_fixes[bugname] = [-1, None, bugname]
-            else:
-                all_fixes[bugid] = [-1, bugid, bugname]
-
-    # Generate lists for all new and old fixes
-    new_fixes = [(mode, bugid, bugname) for dummy, (mode, bugid, bugname) in
-                                            all_fixes.iteritems() if mode > 0]
-    old_fixes = [(mode, bugid, bugname) for dummy, (mode, bugid, bugname) in
-                                            all_fixes.iteritems() if mode <= 0]
-
-    # List of old fixes is not available when releasing a new version
-    if len(old_fixes) == 0:
-        old_fixes = new_fixes
-        new_fixes = []
-
-    # Generate information for current version
-    lines = []
-    if len(new_fixes):
-        lines.append("**Bug fixes and features included in the next upcoming release [%d]:**" % len(new_fixes))
-        lines.append("")
-        for mode, bugid, bugname in sorted(new_fixes, key=lambda x: x[2]):
-            lines.append(_format_bug(mode, bugid, bugname))
-        lines.append("")
-        lines.append("")
-    lines.append("**Bug fixes and features in Wine Staging %s [%d]:**" % (latest_staging_version, len(old_fixes)))
-    lines.append("")
-    lines.append("*Note: The following list only contains features and bug fixes which are not")
-    lines.append("yet available in vanilla Wine. They are removed from the list as soon as they")
-    lines.append("are included upstream. The list also includes features and fixes from previous")
-    lines.append("releases, take a look at the")
-    lines.append("[changelog](https://github.com/wine-compholio/wine-staging/blob/master/staging/changelog)")
-    lines.append("for more details.*")
-    lines.append("")
-    for mode, bugid, bugname in sorted(old_fixes, key=lambda x: x[2]):
-        lines.append(_format_bug(mode, bugid, bugname))
-
-    # Update README.md
-    with open(config.path_template_README_md) as template_fp:
-        template = template_fp.read()
-    with open(config.path_README_md, "w") as fp:
-        fp.write(template.format(fixes="\n".join(lines)))
-
-    # Add changes to git
-    subprocess.call(["git", "add", config.path_README_md])
-
-def wrap_changelog():
-
-    lines = []
-    with open(config.path_changelog) as fp:
-        for line in fp:
-            if line.startswith("  *") and len(line.rstrip("\n")) > 80:
-                wrapped = textwrap.wrap(line[3:].strip(), 80 - 4)
-                lines.append("  * %s\n" % "\n    ".join(wrapped))
-            else:
-                lines.append(line)
-
-    with open(config.path_changelog, "w") as fp:
-        for line in lines:
-            fp.write(line)
-
-    # Add changes to git
-    subprocess.call(["git", "add", config.path_changelog])
-
 
 if __name__ == "__main__":
 
@@ -907,7 +776,7 @@ if __name__ == "__main__":
             raise argparse.ArgumentTypeError("not a valid commit hash")
         return commit
 
-    parser = argparse.ArgumentParser(description="Automatic patch dependency checker and apply script/README.md generator.")
+    parser = argparse.ArgumentParser(description="Automatic patch dependency checker and apply script generator.")
     parser.add_argument('--skip-checks', action='store_true', help="Skip dependency checks")
     parser.add_argument('--commit', type=_check_commit_hash, help="Use given commit hash instead of HEAD")
     parser.add_argument('--sync-bugs', action='store_true', help="Update bugs in bugtracker (requires admin rights)")
@@ -927,23 +796,16 @@ if __name__ == "__main__":
         config.bugtracker_pass = None
 
     try:
-
-        # Get information about Wine and Staging version
-        latest_wine_commit       = _latest_wine_commit(args.commit)
-        latest_staging_version   = _latest_staging_version(only_released=True)
-
-        # Read current and release patches
-        all_patches     = read_patchset()
-        release_patches = read_patchset(revision="v%s" % latest_staging_version)
+        upstream_commit = _upstream_commit(args.commit)
+        all_patches = load_patchsets()
 
         # Check bugzilla
         check_bug_status(all_patches, sync_bugs=args.sync_bugs)
 
         # Update autogenerated files
         generate_ifdefined(all_patches, skip_checks=args.skip_checks)
-        generate_script(all_patches, skip_checks=args.skip_checks)
-        generate_markdown(all_patches, release_patches)
-        wrap_changelog()
+        resolved = generate_apply_order(all_patches, skip_checks=args.skip_checks)
+        generate_script(all_patches, resolved)
 
     except PatchUpdaterError as e:
         print ""
