@@ -91,16 +91,12 @@ class PatchObject(object):
         """Return the full patch as a string."""
         return "".join(chunk for chunk in self.read_chunks())
 
-class _FileReader(object):
-    def __init__(self, filename, content=None):
+class _PatchReader(object):
+    def __init__(self, filename, fp=None):
         self.filename = filename
+        self.fp       = fp if fp is not None else open(filename)
         self.peeked   = None
 
-        if content is not None:
-            self.fp = StringIO(content)
-        else:
-            self.fp = open(filename)
-
     def close(self):
         self.fp.close()
 
@@ -139,6 +135,58 @@ class _FileReader(object):
         tmp, self.peeked = self.peeked, None
         return tmp[1]
 
+    def read_hunk(self):
+        """Read one hunk from a patch file."""
+        line = self.peek()
+        if line is None or not line.startswith("@@ -"):
+            return None
+
+        r = re.match("^@@ -(([0-9]+),)?([0-9]+) \+(([0-9]+),)?([0-9]+) @@", line)
+        if not r: raise PatchParserError("Unable to parse hunk header '%s'." % line)
+        srcpos = max(int(r.group(2)) - 1, 0) if r.group(2) else 0
+        dstpos = max(int(r.group(5)) - 1, 0) if r.group(5) else 0
+        srclines, dstlines = int(r.group(3)), int(r.group(6))
+        if srclines <= 0 and dstlines <= 0:
+            raise PatchParserError("Empty hunk doesn't make sense.")
+        self.read()
+
+        srcdata = []
+        dstdata = []
+
+        try:
+            while srclines > 0 or dstlines > 0:
+                line = self.read().rstrip("\n")
+                if line[0] == " ":
+                    if srclines == 0 or dstlines == 0:
+                        raise PatchParserError("Corrupted patch.")
+                    srcdata.append(line[1:])
+                    dstdata.append(line[1:])
+                    srclines -= 1
+                    dstlines -= 1
+                elif line[0] == "-":
+                    if srclines == 0:
+                        raise PatchParserError("Corrupted patch.")
+                    srcdata.append(line[1:])
+                    srclines -= 1
+                elif line[0] == "+":
+                    if dstlines == 0:
+                        raise PatchParserError("Corrupted patch.")
+                    dstdata.append(line[1:])
+                    dstlines -= 1
+                elif line[0] == "\\":
+                    pass # ignore
+                else:
+                    raise PatchParserError("Unexpected line in hunk.")
+        except IndexError: # triggered by ""[0]
+            raise PatchParserError("Truncated patch.")
+
+        while True:
+            line = self.peek()
+            if line is None or not line.startswith("\\ "): break
+            self.read()
+
+        return (srcpos, srcdata, dstpos, dstdata)
+
 def _read_single_patch(fp, header, oldname=None, newname=None):
     """Internal function to read a single patch from a file."""
 
@@ -150,7 +198,7 @@ def _read_single_patch(fp, header, oldname=None, newname=None):
     # Skip over initial diff --git header
     line = fp.peek()
     if line.startswith("diff --git "):
-        assert fp.read() == line
+        fp.read()
 
     # Read header
     while True:
@@ -192,7 +240,7 @@ def _read_single_patch(fp, header, oldname=None, newname=None):
 
         else:
             break
-        assert fp.read() == line
+        fp.read()
 
     if patch.oldname is None or patch.newname is None:
         raise PatchParserError("Missing old or new name.")
@@ -220,54 +268,15 @@ def _read_single_patch(fp, header, oldname=None, newname=None):
             raise PatchParserError("Stripped old- and new name doesn't match.")
 
     elif line.startswith("@@ -"):
-        while True:
-            line = fp.peek()
-            if line is None or not line.startswith("@@ -"):
-                break
-
-            r = re.match("^@@ -(([0-9]+),)?([0-9]+) \+(([0-9]+),)?([0-9]+) @@", line)
-            if not r: raise PatchParserError("Unable to parse hunk header '%s'." % line)
-            srcpos = max(int(r.group(2)) - 1, 0) if r.group(2) else 0
-            dstpos = max(int(r.group(5)) - 1, 0) if r.group(5) else 0
-            srclines, dstlines = int(r.group(3)), int(r.group(6))
-            if srclines <= 0 and dstlines <= 0:
-                raise PatchParserError("Empty hunk doesn't make sense.")
-            assert fp.read() == line
-
-            try:
-                while srclines > 0 or dstlines > 0:
-                    line = fp.read()[0]
-                    if line == " ":
-                        if srclines == 0 or dstlines == 0:
-                            raise PatchParserError("Corrupted patch.")
-                        srclines -= 1
-                        dstlines -= 1
-                    elif line == "-":
-                        if srclines == 0:
-                            raise PatchParserError("Corrupted patch.")
-                        srclines -= 1
-                    elif line == "+":
-                        if dstlines == 0:
-                            raise PatchParserError("Corrupted patch.")
-                        dstlines -= 1
-                    elif line == "\\":
-                        pass # ignore
-                    else:
-                        raise PatchParserError("Unexpected line in hunk.")
-            except TypeError: # triggered by None[0]
-                raise PatchParserError("Truncated patch.")
-
-            while True:
-                line = fp.peek()
-                if line is None or not line.startswith("\\ "): break
-                assert fp.read() == line
+        while fp.read_hunk() is not None:
+            pass
 
     elif line.rstrip() == "GIT binary patch":
         if patch.oldsha1 is None or patch.newsha1 is None:
             raise PatchParserError("Missing index header, sha1 sums required for binary patch.")
         elif patch.oldname != patch.newname:
             raise PatchParserError("Stripped old- and new name doesn't match for binary patch.")
-        assert fp.read() == line
+        fp.read()
 
         line = fp.read()
         if line is None: raise PatchParserError("Unexpected end of file.")
@@ -320,11 +329,11 @@ def _parse_subject(subject):
     if r is not None: return r.group(1).strip(), 1
     return subject, 1
 
-def read_patch(filename, content=None):
+def read_patch(filename, fp=None):
     """Iterates over all patches contained in a file, and returns PatchObject objects."""
 
     header = {}
-    with _FileReader(filename, content) as fp:
+    with _PatchReader(filename, fp) as fp:
         while True:
             line = fp.peek()
             if line is None:
@@ -333,16 +342,16 @@ def read_patch(filename, content=None):
             elif line.startswith("From: "):
                 header['author'], header['email'] = _parse_author(line[6:])
                 header.pop('signedoffby', None)
-                assert fp.read() == line
+                fp.read()
 
             elif line.startswith("Subject: "):
                 subject = line[9:].rstrip("\r\n")
-                assert fp.read() == line
+                fp.read()
                 while True:
                     line = fp.peek()
                     if not line.startswith(" "): break
                     subject += line.rstrip("\r\n")
-                    assert fp.read() == line
+                    fp.read()
                 subject, revision = _parse_subject(subject)
                 if not subject.endswith("."): subject += "."
                 subject = re.sub('^([^:]*: *)([a-z])', lambda x: "%s%s" %
@@ -354,7 +363,7 @@ def read_patch(filename, content=None):
                 if 'signedoffby' not in header:
                     header['signedoffby'] = []
                 header['signedoffby'].append(_parse_author(line[15:]))
-                assert fp.read() == line
+                fp.read()
 
             elif line.startswith("diff --git "):
                 tmp = line.strip().split(" ")
@@ -368,7 +377,7 @@ def read_patch(filename, content=None):
                 raise PatchParserError("Patch didn't start with a git or diff header.")
 
             else:
-                assert fp.read() == line
+                fp.read()
 
 def apply_patch(original, patchfile, reverse=False, fuzz=2):
     """Apply a patch with optional fuzz - uses the commandline 'patch' utility."""
@@ -503,65 +512,18 @@ def generate_ifdef_patch(original, patched, ifdef):
         lines, split = _preprocess_source(original)
 
         # Parse the created diff file
-        diff.flush()
-        diff.seek(0)
+        fp = _PatchReader(diff.name, diff)
+        fp.seek(0)
 
         # We expect this output format from 'diff', if this is not the case things might go wrong.
-        line = diff.readline()
+        line = fp.read()
         assert line.startswith("--- ")
-        line = diff.readline()
+        line = fp.read()
         assert line.startswith("+++ ")
 
         hunks = []
-        while True:
-            line = diff.readline()
-            if line == "":
-                break
-            if line.startswith("\\ "):
-                continue
-
-            # Parse each hunk, and extract the srclines and dstlines. This algorithm is very
-            # similar to _read_single_patch.
-            if not line.startswith("@@ -"):
-                raise PatchParserError("Unable to parse line '%s'." % line)
-
-            r = re.match("^@@ -(([0-9]+),)?([0-9]+) \+(([0-9]+),)?([0-9]+) @@", line)
-            if not r: raise PatchParserError("Unable to parse hunk header '%s'." % line)
-            srcpos = max(int(r.group(2)) - 1, 0) if r.group(2) else 0
-            dstpos = max(int(r.group(5)) - 1, 0) if r.group(5) else 0
-            srclines, dstlines = int(r.group(3)), int(r.group(6))
-            if srclines <= 0 and dstlines <= 0:
-                raise PatchParserError("Empty hunk doesn't make sense.")
-
-            srcdata = []
-            dstdata = []
-
-            try:
-                while srclines > 0 or dstlines > 0:
-                    line = diff.readline().rstrip("\n")
-                    if line[0] == " ":
-                        if srclines == 0 or dstlines == 0:
-                            raise PatchParserError("Corrupted patch.")
-                        srcdata.append(line[1:])
-                        dstdata.append(line[1:])
-                        srclines -= 1
-                        dstlines -= 1
-                    elif line[0] == "-":
-                        if srclines == 0:
-                            raise PatchParserError("Corrupted patch.")
-                        srcdata.append(line[1:])
-                        srclines -= 1
-                    elif line[0] == "+":
-                        if dstlines == 0:
-                            raise PatchParserError("Corrupted patch.")
-                        dstdata.append(line[1:])
-                        dstlines -= 1
-                    elif line[0] == "\\":
-                        pass # ignore
-                    else:
-                        raise PatchParserError("Unexpected line in hunk.")
-            except IndexError: # triggered by ""[0]
-                raise PatchParserError("Truncated patch.")
+        while fp.peek() is not None:
+            srcpos, srcdata, dstpos, dstdata = fp.read_hunk()
 
             # Ensure that the patch would really apply in practice
             if lines[srcpos:srcpos + len(srcdata)] != srcdata:
@@ -639,9 +601,6 @@ def generate_ifdef_patch(original, patched, ifdef):
                 prev_dstdata.append(lines[prev_endpos])
                 prev_endpos += 1
 
-        # We don't need the diff anymore, all hunks are in memory
-        diff.close()
-
     # Generate resulting file with #ifdefs
     with tempfile.NamedTemporaryFile(mode='w+') as intermediate:
 
@@ -739,40 +698,13 @@ if __name__ == "__main__":
 
     # Basic tests for read_patch()
     class PatchReaderTests(unittest.TestCase):
-        def test_read(self):
-            source = ["From be07df750862699f2515c0ac0ceb7a6c25e9458a Mon Sep 17 00:00:00 2001",
-                      "From: =?UTF-8?q?Author=20Name?= <author@email.com>",
-                      "Subject: [PATCH v3] component: Replace arg1 with arg2.",
-                      "",
-                      "Signed-off-by: =?UTF-8?q?Author=20Name?= <author@email.com>",
-                      "Signed-off-by: Other Developer <other@email.com>",
-                      "---",
-                      " test.txt | 2 +-",
-                      " 1 file changed, 1 insertion(+), 1 deletion(-)",
-                      "",
-                      "diff --git a/test.txt b/test.txt",
-                      "index d54375d..0078e66 100644",
-                      "--- a/test.txt", "+++ b/test.txt",
-                      "@@ -1,7 +1,7 @@",
-                      " line1();", " line2();", " line3();",
-                      "-function(arg1);",
-                      "+function(arg2);",
-                      " line5();", " line6();", " line7();",
-                      "-- ",
-                      "2.7.1"]
-
-            expected = ["diff --git a/test.txt b/test.txt",
-                        "index d54375d..0078e66 100644",
-                        "--- a/test.txt", "+++ b/test.txt",
-                        "@@ -1,7 +1,7 @@",
-                        " line1();", " line2();", " line3();",
-                        "-function(arg1);",
-                        "+function(arg2);",
-                        " line5();", " line6();", " line7();"]
+        def test_simple(self):
+            with open("tests/simple.patch") as fp:
+                source = fp.read().split("\n")
 
             # Test formatted git patch with author and subject
             patchfile = tempfile.NamedTemporaryFile(mode='w+')
-            patchfile.write("\n".join(source + [""]))
+            patchfile.write("\n".join(source))
             patchfile.flush()
 
             patches = list(read_patch(patchfile.name))
@@ -788,13 +720,13 @@ if __name__ == "__main__":
             self.assertEqual(patches[0].modified_file,  "test.txt")
 
             lines = patches[0].read().rstrip("\n").split("\n")
-            self.assertEqual(lines, expected)
+            self.assertEqual(lines, source[10:23])
 
             # Test with git diff
             del source[0:10]
             self.assertTrue(source[0].startswith("diff --git"))
             patchfile = tempfile.NamedTemporaryFile(mode='w+')
-            patchfile.write("\n".join(source + [""]))
+            patchfile.write("\n".join(source))
             patchfile.flush()
 
             patches = list(read_patch(patchfile.name))
@@ -809,13 +741,13 @@ if __name__ == "__main__":
             self.assertEqual(patches[0].modified_file, "test.txt")
 
             lines = patches[0].read().rstrip("\n").split("\n")
-            self.assertEqual(lines, expected)
+            self.assertEqual(lines, source[:13])
 
             # Test with unified diff
             del source[0:2]
             self.assertTrue(source[0].startswith("---"))
             patchfile = tempfile.NamedTemporaryFile(mode='w+')
-            patchfile.write("\n".join(source + [""]))
+            patchfile.write("\n".join(source))
             patchfile.flush()
 
             patches = list(read_patch(patchfile.name))
@@ -829,12 +761,12 @@ if __name__ == "__main__":
             self.assertEqual(patches[0].is_binary,      False)
             self.assertEqual(patches[0].modified_file,  "test.txt")
 
-            del expected[0:2]
             lines = patches[0].read().rstrip("\n").split("\n")
-            self.assertEqual(lines, expected)
+            self.assertEqual(lines, source[:11])
 
             # Test with StringIO buffer
-            patches = list(read_patch("unknown.patch", "\n".join(source + [""])))
+            fp = StringIO("\n".join(source))
+            patches = list(read_patch("unknown.patch", fp))
             self.assertEqual(len(patches), 1)
             self.assertEqual(patches[0].patch_author,   None)
             self.assertEqual(patches[0].patch_email,    None)
@@ -845,6 +777,55 @@ if __name__ == "__main__":
             self.assertEqual(patches[0].is_binary,      False)
             self.assertEqual(patches[0].modified_file,  "test.txt")
 
+        def test_multi(self):
+            with open("tests/multi.patch") as fp:
+                source = fp.read().split("\n")
+
+            patchfile = tempfile.NamedTemporaryFile(mode='w+')
+            patchfile.write("\n".join(source))
+            patchfile.flush()
+
+            patches = list(read_patch(patchfile.name))
+            self.assertEqual(len(patches), 3)
+
+            self.assertEqual(patches[0].patch_author,   "Author Name")
+            self.assertEqual(patches[0].patch_email,    "author@email.com")
+            self.assertEqual(patches[0].patch_subject,  "component: Replace arg1 with arg2.")
+            self.assertEqual(patches[0].patch_revision, 3)
+            self.assertEqual(patches[0].signed_off_by,  [("Author Name", "author@email.com"),
+                                                         ("Other Developer", "other@email.com")])
+            self.assertEqual(patches[0].filename,       patchfile.name)
+            self.assertEqual(patches[0].is_binary,      False)
+            self.assertEqual(patches[0].modified_file,  "other_test.txt")
+
+            lines = patches[0].read().rstrip("\n").split("\n")
+            self.assertEqual(lines, source[11:24])
+
+            self.assertEqual(patches[1].patch_author,   "Author Name")
+            self.assertEqual(patches[1].patch_email,    "author@email.com")
+            self.assertEqual(patches[1].patch_subject,  "component: Replace arg1 with arg2.")
+            self.assertEqual(patches[1].patch_revision, 3)
+            self.assertEqual(patches[1].signed_off_by,  [("Author Name", "author@email.com"),
+                                                         ("Other Developer", "other@email.com")])
+            self.assertEqual(patches[1].filename,       patchfile.name)
+            self.assertEqual(patches[1].is_binary,      False)
+            self.assertEqual(patches[1].modified_file,  "test.txt")
+
+            lines = patches[1].read().rstrip("\n").split("\n")
+            self.assertEqual(lines, source[24:46])
+
+            self.assertEqual(patches[2].patch_author,   "Other Developer")
+            self.assertEqual(patches[2].patch_email,    "other@email.com")
+            self.assertEqual(patches[2].patch_subject,  "component: Replace arg2 with arg3.")
+            self.assertEqual(patches[2].patch_revision, 4)
+            self.assertEqual(patches[2].signed_off_by,  [("Other Developer", "other@email.com")])
+            self.assertEqual(patches[2].filename,       patchfile.name)
+            self.assertEqual(patches[2].is_binary,      False)
+            self.assertEqual(patches[2].modified_file,  "test.txt")
+
+            lines = patches[2].read().rstrip("\n").split("\n")
+            self.assertEqual(lines, source[58:71])
+
     # Basic tests for apply_patch()
     class PatchApplyTests(unittest.TestCase):
         def test_apply(self):
